from typer import Typer, Option, Argument, echo, colors, style
from typer.testing import CliRunner
from pylibup.types import *
from pylibup.utils import get_logger

logger = get_logger()


"""
name: Optional[str] = Default(None),
cls: Optional[Type[click.Command]] = Default(None),
invoke_without_command: bool = Default(False),
no_args_is_help: bool = Default(False),
subcommand_metavar: Optional[str] = Default(None),
chain: bool = Default(False),
result_callback: Optional[Callable[..., Any]] = Default(None),
# Command
context_settings: Optional[Dict[Any, Any]] = Default(None),
callback: Optional[Callable[..., Any]] = Default(None),
help: Optional[str] = Default(None),
epilog: Optional[str] = Default(None),
short_help: Optional[str] = Default(None),
options_metavar: str = Default("[OPTIONS]"),
add_help_option: bool = Default(True),
hidden: bool = Default(False),
deprecated: bool = Default(False),
add_completion: bool = True,
"""

def createCli(name: str, invoke_without_command: bool = False, chain: bool = False, result_callback: Optional[Callable[..., Any]] = None, **kwargs):
    return Typer(name=name, invoke_without_command=invoke_without_command, chain=chain, result_callback=result_callback, **kwargs)

baseCli = Typer()

__all__ = [
    'List',
    'Optional',
    'Typer',
    'Option',
    'Callable',
    'Argument',
    'echo',
    'style',
    'colors',
    'createCli',
    'CliRunner',
    'logger',
]


import requests
from git import Repo
from github import Github
from jinja2 import Template

from .envs import envToStr
from .types import *
from .utils import get_logger, to_path, Path, exec_shell
from .serializers import Yaml, Json, Base
from .static import *


logger = get_logger()

class BaseCls(BaseModel):
    class Config:
        arbitrary_types_allowed = True
    
    def get(self, name, default: Any = None):
        return getattr(self, name, default)


class PylibStructure(BaseCls):
    modules: Optional[List[str]] = []

class PylibDockerBuildOptions(BaseCls):
    app_name: Optional[str]
    require_ecr: Optional[bool] = False
    ecr_options: Optional[Dict[str, Any]] = {}
    docker_options: Optional[Dict[str, Any]] = {}

class PylibGithubWorkflows(BaseCls):
    pypi_publish: Optional[bool] = True
    docker_build: Optional[bool] = False
    docker_build_options: Optional[PylibDockerBuildOptions] = Field(default=PylibDockerBuildOptions)
    

class PylibOptions(BaseCls):
    default_branch: Optional[str] = 'main'
    include_init: Optional[bool] = True
    include_app: Optional[bool] = False
    include_dockerfile: Optional[bool] = False
    include_buildscript: Optional[bool] = True
    include_reqtext: Optional[bool] = True
    private: Optional[bool] = True


class PylibConfigData(BaseCls):
    setup: Optional[Dict[str, Any]]
    repo: Optional[str]
    readme_text: Optional[str]
    project_description: Optional[str]
    gitignores: Optional[List[str]]
    structure: Optional[PylibStructure]
    secrets: Optional[Dict[str, Any]]
    options: Optional[PylibOptions] = Field(default = PylibOptions)
    workflows: Optional[PylibGithubWorkflows] = Field(default = PylibGithubWorkflows)

    @property
    def opt(self) -> PylibOptions: return self.options
    
    @property
    def wkflw(self) -> PylibGithubWorkflows: return self.workflows
    
    @property
    def libname(self) -> str:
        if self.setup.get('lib_name'): return self.setup['lib_name']
        if self.setup.get('pkg_name'): return self.setup['pkg_name']
        return self.repo_name or ''


    @property
    def description(self):
        if self.project_description: return self.project_description
        if self.setup.get('description'): return self.setup['description']
        return None

    @property
    def user_repo(self):
        if self.repo: return self.repo.split('/', 1)[0].strip()
        if self.setup and self.setup.get('git_repo'): return self.setup['git_repo']
        return None
    
    @property
    def repo_name(self):
        if self.repo: return self.repo.split('/', 1)[-1].strip()
        if self.setup: 
            if self.setup.get('pkg_name'): return self.setup['pkg_name']
            if self.setup.get('lib_name'): return self.setup['lib_name']
        return None

    @property
    def repo_path(self):
        if not self.user_repo and not self.repo_name: return None
        return f'{self.user_repo}/{self.repo_name}'
    
    @property
    def repo_url(self):
        return f'https://github.com/{self.repo_path}.git'
    
    @property
    def tmpl_setup_py(self):
        if not self.setup: return None
        tmpl = Template(setup_py_template)
        return tmpl.render(self.setup)
    
    @property
    def tmpl_requirements_txt(self):
        if not self.opt.include_reqtext: return None
        tmpl = Template(install_requirements_template)
        return tmpl.render(self.setup)

    @property
    def tmpl_readme_md(self):
        if not self.readme_text: return None
        tmpl = Template(readme_template)
        readme_data = self.setup or {}
        readme_data['readme_text'] = self.readme_text
        return tmpl.render(readme_data)
    
    @property
    def tmpl_gitignore(self):
        if not self.gitignores: return None
        tmpl = Template(gitignores_template)
        data = {'gitignore': self.gitignores}
        return tmpl.render(data)
    
    @property
    def tmpl_workflows_enabled(self):
        return bool(self.wkflw.docker_build or self.wkflw.pypi_publish)

    @property
    def tmpl_github_action_pypi_publish(self):
        if not self.wkflw.pypi_publish: return None
        return github_action_template_pypi_publish    
    
    @property
    def tmpl_github_action_docker_build(self):
        if not self.wkflw.docker_build: return None
        tmpl = Template(github_action_template_docker_build)
        data = {
            'app_name': self.wkflw.docker_build_options.app_name or self.setup.get('lib_name', self.setup.get('pkg_name')), 
            'require_ecr': self.wkflw.docker_build_options.require_ecr,
            'ecr_options': self.wkflw.docker_build_options.ecr_options,
            'docker_options': self.wkflw.docker_build_options.docker_options,
        }
        return tmpl.render(data)

    @property
    def tmpl_build_sh(self):
        if not self.opt.include_buildscript: return None
        return build_sh_template
    
    @property
    def tmpl_init_py(self):
        if not self.opt.include_init: return None
        tmpl = Template(pyinit_template)
        data = {'modules': self.structure.modules}
        return tmpl.render(data)
    
    @property
    def tmpl_dockerfile_app(self):
        if not self.opt.include_app and not self.opt.include_dockerfile: return None
        return dockerfile_fastapi_template
    
    @property
    def needs_ipyirc(self):
        return self.wkflw.pypi_publish
    
    def should_add_to_commit(self, filename: str):
        return not any(i in filename or filename in i for i in self.gitignores)

    def get_secrets(self):
        data = {}
        for key, val in self.secrets:
            if val:
                if isinstance(val, str): data[key] = val
                if isinstance(val, dict) and val.get('from'):
                    data[key] = envToStr(val['from'], envToStr(key))
            else: data[key] = envToStr(envToStr(key))
        return data


class PylibConfig:
    def __init__(self, github: Github, github_token: str, config_file: str, project_name: str, project_dir: str = None, *args, **kwargs):
        self.github = github
        self.github_token = github_token
        self.config_file = config_file
        self.repo_files = []
        self.set_working_project(project_name, project_dir)
        self.configfile_data = self.load_config_file(self.config_file)
        self.config = self.load_config_data(self.configfile_data)
        self.setup_repo()

    def setup_repo(self):
        if not self.working_dir.joinpath('.git').exists(): self.repo = Repo.init(self.working_dir, bare=False, initial_branch=self.config.opt.default_branch)
        else: self.repo = Repo(self.working_dir, search_parent_directories=True)

    def set_working_project(self, project_name: str = None, project_dir: str = None):
        self.current_dir = to_path(project_dir) if project_dir else Path.cwd()
        if not project_name: project_name = self.current_dir.stem
        self.working_dir = self.current_dir.joinpath(project_name) if project_name not in self.current_dir.as_posix() else self.current_dir
        self.project_name = project_name
        if not self.working_dir.exists():
            self.working_dir.mkdir(parents=True)
        self.workflow_dir = self.working_dir.joinpath('.github/workflows')
        self.app_dir = self.working_dir.joinpath('app')
        if not self.config_file:
            if self.working_dir.joinpath('metadata.yaml').exists():
                self.config_file = self.working_dir.joinpath('metadata.yaml')
            elif self.current_dir.joinpath('metadata.yaml').exists():
                self.config_file = self.current_dir.joinpath('metadata.yaml')
            assert self.config_file, 'No specified config file'

    @classmethod
    def load_config_file(cls, config_file: str):
        config_file = to_path(config_file)
        if config_file.suffix == '.json': loader = Json.loads
        elif config_file.suffix == '.yaml': loader = Yaml.loads
        return loader(config_file.read_text())
    
    @classmethod
    def load_config_data(cls, config_data: Dict[str, Any]) -> PylibConfigData:
        return PylibConfigData(**config_data)
    
    @property
    def github_username(self) -> str:
        return self.github.get_user().login
    
    @property
    def github_repo_user(self) -> str:
        if self.config.user_repo: return self.config.user_repo
        return self.github_username

    @property
    def github_repo_name(self) -> str:
        if self.config.repo_name: return self.config.repo_name
        return self.project_name

    @property
    def github_repo_path(self):
        if self.config.repo_path: return self.config.repo_path
        return f'{self.github_repo_user}/{self.github_repo_name}'
    
    @property
    def user_authentication(self):
        return Base.b64_encode(f'{self.github_username}:{self.github_token}')

    @property
    def github_headers(self):
        return {'Accept': 'application/json, application/vnd.github.v3+json', 'Content-Type': 'application/json', 'Authorization': f'Basic {self.user_authentication}'}

    def get_github_repo(self):
        return requests.get(f'https://api.github.com/repos/{self.github_repo_path}', headers=self.github_headers)

    def create_github_repo(self, **kwargs):
        if self.repo_exists: return
        data = { 
            'name': self.github_repo_name,
            'private': self.config.opt.private,
            'description': self.config.description,
            'default_branch': self.config.opt.default_branch,
        }
        if kwargs: data.update(kwargs)
        return requests.post(url = 'https://api.github.com/user/repos', headers=self.github_headers, json=data)

    @property
    def repo_exists(self) -> bool:
        return bool(self.get_github_repo().status_code < 399)
    
    def publish_repo(self):
        if not self.repo_exists:
            self.create_github_repo()
            exec_shell(f'cd {self.working_dir} && git remote add origin {self.config.repo_url} && git branch -M {self.config.opt.default_branch}')
        exec_shell(f'cd {self.working_dir} && git push -u origin {self.config.opt.default_branch}')


    def build_tmpl(self, tmpl_data: Union[str, Any], filename: str, overwrite: bool = False, add_to_commit: bool = True):
        if tmpl_data:
            tmpl_file = self.working_dir.joinpath(filename)
            if tmpl_file.exists() and not overwrite: pass
            logger(f'Building: {filename}')
            tmpl_file.write_text(tmpl_data)
            if add_to_commit:
                self.repo_files.append(tmpl_file.as_posix())

    def build_base(self, overwrite: bool = False, *args, **kwargs):
        self.build_tmpl(tmpl_data = self.config.tmpl_setup_py,  filename = 'setup.py',  overwrite=overwrite)
        self.build_tmpl(tmpl_data = self.config.tmpl_build_sh,  filename = 'build.sh',  overwrite=overwrite, add_to_commit = self.config.should_add_to_commit('build.sh'))
        self.build_tmpl(tmpl_data = self.config.tmpl_requirements_txt,  filename = 'requirements.txt', overwrite=overwrite)
        self.build_tmpl(tmpl_data = self.config.tmpl_readme_md,  filename = 'README.md',  overwrite=overwrite)
        self.build_tmpl(tmpl_data = self.config.tmpl_gitignore,  filename = '.gitignore',  overwrite=overwrite)

    def build_pylib_structure(self, overwrite: bool = False, *args, **kwargs):
        if not self.config.structure: return
        logger('Setting up Pylib structure')
        pydir = self.working_dir.joinpath(self.config.libname)
        pydir.mkdir(parents=True, exist_ok=True)
        for module in self.config.structure.modules:
            tmpl_file = pydir.joinpath(f'{module}.py')
            if tmpl_file.exists() and not overwrite: pass
            logger(f'Adding {self.config.libname}/{module}.py')
            tmpl_file.touch(exist_ok=True)
            self.repo_files.append(tmpl_file.as_posix())
        
        if self.config.opt.include_init:
            tmpl_file = pydir.joinpath('__init__.py')
            if tmpl_file.exists() and not overwrite: pass
            tmpl_file.write_text(self.config.tmpl_init_py)
            self.repo_files.append(tmpl_file.as_posix())
    
    def build_github_workflows(self, overwrite: bool = False, *args, **kwargs):
        if not self.config.tmpl_workflows_enabled: return
        logger('Setting up Github Workflows')
        self.workflow_dir.mkdir(parents=True, exist_ok=True)

        if self.config.tmpl_github_action_pypi_publish:
            tmpl_file = self.workflow_dir.joinpath('python-publish.yaml')
            if tmpl_file.exists() and not overwrite: pass
            logger('Building: .github/workflows/python-publish.yaml')
            tmpl_file.write_text(self.config.tmpl_github_action_pypi_publish)
            self.repo_files.append(tmpl_file.as_posix())
        
        if self.config.tmpl_github_action_docker_build:
            tmpl_file = self.workflow_dir.joinpath('docker-build.yaml')
            if tmpl_file.exists() and not overwrite: pass
            logger('Building: .github/workflows/docker-build.yaml')
            tmpl_file.write_text(self.config.tmpl_github_action_docker_build)
            self.repo_files.append(tmpl_file.as_posix())
    
    def build_docker_app(self, overwrite: bool = False, *args, **kwargs):
        if not self.config.opt.include_app: return
        logger('Setting up AppDir')
        self.app_dir.mkdir(parents=True, exist_ok=True)
        for appfile in ['__init__', 'config', 'client', 'classes', 'routez', 'utils']:
            tmpl_file = self.app_dir.joinpath(f'{appfile}.py')
            if tmpl_file.exists() and not overwrite: pass
            logger(f'Adding app/{appfile}.py')
            tmpl_file.touch(exist_ok=True)
            self.repo_files.append(tmpl_file.as_posix())

        if self.config.tmpl_dockerfile_app:
            tmpl_file = self.working_dir.joinpath('Dockerfile')
            if tmpl_file.exists() and not overwrite: pass
            logger('Adding Dockerfile for App')
            tmpl_file.write_text(self.config.tmpl_dockerfile_app)
            self.repo_files.append(tmpl_file.as_posix())

    
    def build(self, commit_msg: str = 'Initialize', overwrite: bool = False, auto_publish: bool = False, *args, **kwargs):
        logger.info('====================================================')
        logger.info(f'Building Repo: {self.project_name} @ {self.config.opt.default_branch}')
        logger.info('====================================================')
        self.build_base(overwrite=overwrite, *args, **kwargs)
        self.build_pylib_structure(overwrite=overwrite, *args, **kwargs)
        self.build_github_workflows(overwrite=overwrite, *args, **kwargs)
        self.build_docker_app(overwrite=overwrite, *args, **kwargs)
        if self.repo_files:
            logger.info(f'Adding {len(self.repo_files)} Files to Git Index')
            self.repo.index.add(self.repo_files)
        logger.info(f'Adding Commit: {commit_msg}')
        self.repo.index.commit(commit_msg)
        if auto_publish:
            logger.info(f'Publishing {self.project_name}')
            self.publish_repo()
        logger('Completed Pylib Build. Have fun!')
    
    def publish_repo(self, commit_msg: str = None):
        if self.repo_exists: logger.warn(f'Repo already exists. Skipping Publishing: {self.project_name}')
        else: logger.info(f'Publishing {self.project_name}')
        if commit_msg: 
            logger.info(f'Adding Commit: {commit_msg}')
            self.repo.index.commit(commit_msg)
        self.publish_repo()


def get_metadata_template(github: Github, name: str, repo_user: str = None, private: bool = True, **kwargs):
    metadata = default_pylib_metadata.copy()
    #if kwargs: metadata.update(kwargs)
    caller = github.get_user()
    if not repo_user: repo_user = caller.login
    metadata['repo'] = f'{repo_user}/{name}'
    #metadata['options']['private'] = private
    metadata['setup'].update({
        'author': caller.name or repo_user,
        'email': caller.email,
        'git_repo': f'{repo_user}/{name}',
        'pkg_name': name,
        'lib_name': kwargs.get('lib_name', name)
    })
    metadata['project_description'] = kwargs.get('description', '')
    metadata['readme_text'] = kwargs.get('readme_text', '')
    if kwargs.get('secrets'):
        metadata['secrets'] = kwargs['secrets']
    #logger.info(metadata)
    return Yaml.dumps(metadata)




    



        


        









setup_py_template = """
import os
import sys
from pathlib import Path
from setuptools import setup, find_packages

{% if require_py3 %}
if sys.version_info.major != 3:
    raise RuntimeError("This package requires Python 3+")
{% endif %}

version = '{{ pkg_version }}'
pkg_name = '{{ pkg_name }}'
gitrepo = '{{ git_repo }}/{{ pkg_name }}'
root = Path(__file__).parent

requirements = [
    {%- for item in requirements %}
    '{{ item }}',
    {%- endfor %}
]

args = {
    'packages': find_packages(include = ['{{ lib_name }}', '{{ lib_name }}.*']),
    'install_requires': requirements,
    'long_description': root.joinpath('README.md').read_text(encoding='utf-8'),
    {%- if require_py3_version %}
    'python_requires': '>={{ require_py3_version }}',
    {%- endif %}
    {%- if kwargs %}
    {%- for key, value in kwargs|dictsort %}
    '{{ key }}': {{ value }},
    {%- endfor %}
    {%- endif %}
    {%- if include_pkg_files %}
    'include_package_data': True,
    {%- endif %}
    {%- if data_files %}
    {%- for key, value in data_files|dictsort %}
    '{{ key }}': {{ value }},
    {%- endfor %}
    {%- endif %}
    'entry_points': {
        {%- if cli_cmds %}
        'console_scripts': [
        {%- for item in cli_cmds %}
            '{{ item }}',
        {%- endfor %}
        ],
        {%- endif %}
        {%- if entry_points %}
        {%- for key, value in entry_points|dictsort %}
        '{{ key }}': {{ value }},
        {%- endfor %}
        {%- endif %}
    }
}

setup(
    name=pkg_name,
    version=version,
    url='https://github.com/{{ git_repo }}/{{ pkg_name }}',
    license='MIT Style',
    description='{{ description }}',
    author='{{ author }}',
    author_email='{{ email }}',
    long_description_content_type="text/markdown",
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: {% if require_py3_version %}{{ require_py3_version }}{% endif %}',
        'Topic :: Software Development :: Libraries',
    ],
    **args
)
"""

install_requirements_template = """
{%- for item in requirements %}
{{ item }}
{%- endfor %}
"""

gitignores_template = """
{%- for item in gitignore %}
{{ item }}
{%- endfor %}
"""

pyinit_template = """
{%- for item in modules %}
from . import {{ item }}
{%- endfor %}
"""

readme_template = """
# {{ pkg_name }}
 {{ description }}

{%- if readme_text %}
---
{{ readme_text }}
{%- endif %}

---
## Quickstart

```bash
# From pypi
pip install --upgrade {{ pkg_name }}

# From source
pip install --upgrade git+https://github.com/{{ git_repo }}/{{ pkg_name }}
```

## Usage

```python
import {{ lib_name }}

## Do something here

```

---

## Libraries & Dependencies

{%- if require_py3_version %}
- `Python {{ require_py3_version }}`
{%- endif %}
{%- for item in requirements %}
- `{{ item }}'`
{% endfor %}

---

This Python Library has been autogenerated with [pylibup](https://github.com/trisongz/pylibup).

"""

build_sh_template = """
#!/bin/bash

## Autogenerated from Pylibup

MODE=$1

if [[ "$MODE" == "prod" || "$MODE" == "dist" ]]; then
    UPLOAD="dist"
else
    UPLOAD="test"
fi

rm -r dist/*
rm -r build/*

python setup.py sdist bdist_wheel

if [[ "$UPLOAD" == "dist "]]; then
    echo "Uploading to Dist Pypi"
    python -m twine upload --repository pypi dist/*
else
    echo "Uploading to Test Pypi"
    python -m twine upload --repository testpypi dist/*
fi
"""


github_action_template = """
## Autogenerated from Pylibup

name: Upload Python Package
on:
  push:
    paths:
      - 'setup.py'
  release:
    types: [created]
jobs:
  deploy:
    runs-on: ubuntu-latest
    steps:
    - uses: actions/checkout@v2
    - name: Set up Python
      uses: actions/setup-python@v2
      with:
        python-version: '3.x'
    - name: Install dependencies
      run: |
        python -m pip install --upgrade pip
        pip install build
    - name: Build package
      run: python -m build
    - name: Publish package
      uses: pypa/gh-action-pypi-publish@release/v1
      with:
        user: __token__
        password: ${{ secrets.PYPI_API_TOKEN }}
"""



github_action_template_pypi_publish = """
## Autogenerated from Pylibup

name: Upload Python Package
on:
  push:
    paths:
      - 'setup.py'
  release:
    types: [created]
jobs:
  deploy:
    runs-on: ubuntu-latest
    steps:
    - uses: actions/checkout@v2
    - name: Set up Python
      uses: actions/setup-python@v2
      with:
        python-version: '3.x'
    - name: Install dependencies
      run: |
        python -m pip install --upgrade pip
        pip install build
    - name: Build package
      run: python -m build
    - name: Publish package
      uses: pypa/gh-action-pypi-publish@release/v1
      with:
        user: __token__
        password: ${{ secrets.PYPI_API_TOKEN }}
"""

github_action_template_docker_build = """
## Autogenerated from Pylibup

name: Build {{ app_name }} Docker Image
on:
  push:
    paths:
      - 'app/**/**'
      - '{{ lib_name }}/**'
      - 'Dockerfile'
      - 'requirements.txt'
      - '.github/workflows/docker-build.yaml'

env:
  {%- if require_ecr %}
  IMG_REPO: {{ ecr_options.repo }}
  {% else %}
  IMG_REPO: {{ docker_options.repo }}
  {%- endif %}

jobs:
  build-latest-docker-image:
    runs-on: ubuntu-latest
    permissions:
      contents: read
      packages: write
    steps:
      - name: Checkout repository
        uses: actions/checkout@v2

      - name: Set up Docker Buildx
        id: buildx
        uses: docker/setup-buildx-action@v1
      {%- if require_ecr %}
      - name: Configure AWS credentials
        uses: aws-actions/configure-aws-credentials@v1
        with:
          aws-access-key-id: ${{ secrets.AWS_ACCESS_KEY_ID }}
          aws-secret-access-key: ${{ secrets.AWS_SECRET_ACCESS_KEY }}
          aws-region: ${{ secrets.AWS_REGION }}

      - name: Login to Amazon ECR
        id: login-ecr
        uses: aws-actions/amazon-ecr-login@v1
      
      - name: Create ECR Repo if not Exists
        uses: int128/create-ecr-repository-action@v1
        id: ecr
        with:
          repository: ${{ env.IMG_REPO }}
      - name: Build and Push Docker Image: {{ app_name }}
        uses: docker/build-push-action@v2
        with:
          file: Dockerfile
          platforms: linux/amd64
          push: true
          tags: |
            ${{ steps.ecr.outputs.repository-uri }}:latest
          cache-from: type=gha
          cache-to: type=gha,mode=max
      {% else %}
      - name: Build and Push Docker Image: {{ app_name }}
        uses: docker/build-push-action@v2
        with:
          file: Dockerfile
          platforms: linux/amd64
          push: true
          tags: |
            ${{ env.IMG_REPO }}:latest
          cache-from: type=gha
          cache-to: type=gha,mode=max
      {%- endif %}
"""

dockerfile_fastapi_template = """
## Autogenerated from Pylibup

FROM tiangolo/uvicorn-gunicorn-fastapi:python3.9

COPY ./requirements.txt /app/requirements.txt

RUN pip install --no-cache-dir --upgrade -r /app/requirements.txt

COPY ./app /app

WORKDIR /app

ENV PYTHONPATH=/app:$PYTHONPATH
"""


pylib_metadata_template = """
# Autogenerated by Pylib

{%- if setup %}
setup:
  {%- for key, value in setup %}
  {{ key }}: {{ value }}
  {%- endfor %}
{%- endif %}

{%- if repo %}
repo: {{ repo }}
{%- endif %}

{%- if readme_text %}
readme_text: {{ readme_text }}
{%- endif %}

{%- if project_description %}
project_description: {{ project_description }}
{%- endif %}

{%- if gitignores %}
gitignores: 
  {%- for item in gitignores %}
  - '{{ item }}'
  {%- endfor %}
{%- endif %}

{%- if structure %}
structure:
  {%- if structure.modules %}
  modules:
  {%- for item in structure.modules %}
  - {{ item }}
  {%- endfor %}
  {%- endif %}

{%- endif %}
{%- if secrets %}
secrets:
  {%- for key, value in secrets %}
  {{ key }}: {{ value }}
  {%- endfor %}
{%- endif %}
{%- if options %}
options:
  {%- for key, value in options|dictsort %}
  {{ key }}: {{ value }}
  {%- endfor %}
{%- endif %}
{%- if workflows %}
workflows:
  {%- for key, value in workflows %}
  {{ key }}: {{ value }}
  {%- endfor %}
{%- endif %}

"""

default_metadata_setup = {
  'author': 'Tri Songz',
  'email': '',
  'git_repo': '',
  'pkg_version': '0.0.0a',
  'pkg_name': 'pylibup',
  'lib_name': 'pylibup',
  'require_py3': True,
  'require_py3_version': 3.7,
  'requirements': ['lazycls', 'pylogz'],
  'kwargs': {},
  'cli_cmds': [],
}

default_metadata_gitignores = [
  'cache*',
  '*.DS_Store',
  'tests*',
  '__pycache__*',
  '*logs',
  '*dist',
  '*build',
  '**build.sh',
  '*test.py',
  '*.egg-info*',
  '*.vscode',
  '**.ipynb',
  '**meta.yaml',
  '**metadata.yaml',
  '**state.yaml'
]

default_metadata_modules = [
  'classes',
  'client',
  'config',
  'utils',
]

default_metadata_secrets = {
  'AWS_ACCESS_KEY_ID': {'from': 'AWS_ACCESS_KEY_ID_SVC_ACCOUNT'},
  'AWS_SECRET_ACCESS_KEY': {'from': 'AWS_SECRET_ACCESS_KEY_SVC_ACCOUNT'},
  'AWS_REGION': 'us-east-1',
}
default_metadata_options = {
  'default_branch': 'main',
  'include_init': True,
  'include_app': True,
  'include_dockerfile': True,
  'include_buildscript': True,
  'include_reqtext': True,
  'private': True,
}

default_metadata_dockerbuild_options = {
  'app_name':'',
  'require_ecr': True,
  'ecr_options': {
    'img_repo': ''
  },
  'docker_options':{
    'img_repo': ''
  }
}

default_metadata_workflows = {
  'pypi_publish': True,
  'docker_build': False,
  'docker_build_options': default_metadata_dockerbuild_options
}

default_pylib_metadata = {
  'setup': default_metadata_setup,
  'repo': '',
  'readme_text': '',
  'project_description': '',
  'gitignores': default_metadata_gitignores,
  'structure': {'modules': default_metadata_modules},
  'secrets': default_metadata_secrets,
  'options': default_metadata_options,
  'workflows': default_metadata_workflows
}




from pathlib import Path
from pylibup.client import PylibClient, Github
from pylibup.cli.base import *
from pylibup.serializers import Yaml
from pylibup.utils import to_path, get_parent_path, exec_shell
from typing import List
import typer

repoCli = createCli(name = 'repo')
stateCli = createCli(name = 'state')

def get_cwd(*paths, posix: bool = True):
    if not paths:
        if posix: return Path.cwd().as_posix()
        return Path.cwd()
    if posix: return Path.cwd().joinpath(*paths).as_posix()
    return Path.cwd().joinpath(*paths)

statefile = get_cwd('.pylibstate.yaml', posix=False)
globalstate_dir = get_parent_path(__file__).joinpath('.pylibstate')
globalstate_dir.mkdir(exist_ok=True)
globalstatefile = globalstate_dir.joinpath('state.yaml')
globalstate_ignore_keys = {'name', 'commit_msg', 'project_name', 'project_dir'}

def load_global_state():
    if globalstatefile.exists(): return Yaml.loads(globalstatefile.read_text())
    return {}

def load_state():
    if statefile.exists(): return Yaml.loads(statefile.read_text())
    return {}

def load_merged_states():
    # global first, local overrides
    globaldata = load_global_state()
    localdata = load_state()
    if localdata: globaldata.update(localdata)
    return globaldata

def save_global_state(overwrite_state: bool = False, **kwargs):
    if globalstatefile.exists():
        data = load_global_state()
        if overwrite_state:
            data.update({k:v for k,v in kwargs.items() if v})
            kwargs = data
        else:
            kwargs.update({k:v for k,v in data.items() if v})
    kwargs = {k:v for k,v in kwargs.items() if k not in globalstate_ignore_keys and v is not None}
    globalstatefile.write_text(Yaml.dumps(kwargs))

def save_state(overwrite_state: bool = False, **kwargs):
    if statefile.exists():
        data = load_state()
        if overwrite_state:
            data.update({k:v for k,v in kwargs.items() if v})
            kwargs = data
        else:
            kwargs.update({k:v for k,v in data.items() if v})
    statefile.write_text(Yaml.dumps(kwargs))


@repoCli.command('init')
def init_new_repo(
    name: Optional[str] = Argument(None),
    project_dir: Optional[str] = Argument(get_cwd()),
    repo_user: Optional[str] = Argument(None),
    github_token: Optional[str] = Option("", envvar="GITHUB_TOKEN"),
    private: bool = Option(True, "--public"),
    overwrite: bool = Option(False),
    overwrite_state: bool = Option(False),
    ):
    state = load_merged_states()
    github_token = github_token or state.get('github_token', '')
    client = PylibClient(github_token = github_token)
    try:
        client.init(project_dir = project_dir, name = name, repo_user = repo_user, private = private, overwrite = overwrite)
        save_state(name = name, project_dir = project_dir, repo_user = repo_user, github_token = github_token, private = private, overwrite = overwrite, overwrite_state = overwrite_state)
    except Exception as e:
        logger.error(e)

@repoCli.command('build')
def build_new_repo(
    config_file: Optional[str] = Argument(get_cwd('metadata.yaml')),
    name: Optional[str] = Argument(None),
    project_dir: Optional[str] = Argument(get_cwd()),
    github_token: Optional[str] = Option("", envvar="GITHUB_TOKEN"),
    pypirc_path: Optional[str] = Option("~/.pypirc", envvar="PYPIRC_PATH"),
    commit_msg: Optional[str] = Option("Initialize"),
    auto_publish: bool = Option(False),
    overwrite: bool = Option(False),
    overwrite_state: bool = Option(False),
    ):
    state = load_merged_states()
    github_token = github_token or state.get('github_token', '')
    pypirc_path = state.get('pypirc_path', pypirc_path)
    config_file = config_file or state.get('config_file')
    project_dir = project_dir or state.get('project_dir')
    client = PylibClient(github_token = github_token, pyirc_path = pypirc_path)
    try:
        client.build(config_file = config_file, project_name = name, project_dir = project_dir, commit_msg = commit_msg, auto_publish = auto_publish, overwrite = overwrite)
        save_state(github_token = github_token, pyirc_path = pypirc_path, config_file = config_file, project_name = name, project_dir = project_dir, commit_msg = commit_msg, auto_publish = auto_publish, overwrite = overwrite, overwrite_state = overwrite_state)
    except Exception as e:
        logger.error(e)


@repoCli.command('cleanup')
def cleanup_repo(
    force: bool = Option(False),
    keep_dir: bool = Option(True),
    ):
    state = load_merged_states()
    project_dir = state.get('project_dir', get_cwd())
    logger.info(f'Planning to remove everything in {project_dir}')
    if not force: force = typer.confirm("Are you sure you want to delete everything in this repo? There is no going back.", abort=True)
    if keep_dir:
        logger.info(f'Removing all files in {project_dir}/*')
        exec_shell(f'rm -rf {project_dir}/*')
    else:
        logger.info(f'Removing directory {project_dir}')
        exec_shell(f'rm -rf {project_dir}')


@repoCli.command('publish')
def publish_new_repo(
    config_file: Optional[str] = Argument(get_cwd('metadata.yaml')),
    github_token: Optional[str] = Option("", envvar="GITHUB_TOKEN"), 
    pypirc_path: Optional[str] = Option("~/.pypirc", envvar="PYPIRC_PATH"),
    commit_msg: Optional[str] = Option("Initialize"),
    overwrite_state: bool = Option(False),
    ):
    state = load_merged_states()
    github_token = github_token or state.get('github_token', '')
    pypirc_path = state.get('pypirc_path', pypirc_path)
    config_file = config_file or state.get('config_file')
    project_dir = state.get('project_dir')
    project_name = state.get('project_name')
    client = PylibClient(github_token = github_token, pyirc_path = pypirc_path)
    try:
        client.publish(commit_msg= commit_msg, config_file = config_file, project_dir = project_dir, project_name = project_name)
        save_state(github_token = github_token, pyirc_path = pypirc_path, config_file = config_file, project_name = project_name, project_dir = project_dir, commit_msg = commit_msg, overwrite_state = overwrite_state)
    except Exception as e:
        logger.error(e)

@repoCli.command('push')
def push_to_repo(
    commit: Optional[str] = Argument("Updating"),
    branch: Optional[str] = Option("main"),
    add_files: bool = Option(True, '--no-add'),
    ):
    cmd = f'cd {get_cwd()} && '
    if add_files: cmd += 'git add . && '
    cmd += f'git commit -m "{commit}" && git push -u origin {branch}'
    exec_shell(cmd)


@repoCli.command('release')
def publish_release(
    tag: Optional[str] = Option("v0.0.1"),
    tag_message: Optional[str] = Option("Stable Release"),
    release_name: Optional[str] = Option("Release v0.0.1"),
    release_message: Optional[str] = Option("Stable Release"),
    draft: bool = Option(False),
    prerelease: bool = Option(False),
    branch: Optional[str] = Option("main"),
    push_first: bool = Option(True, '--no-push'),
    github_token: Optional[str] = Option("", envvar="GITHUB_TOKEN")
    ):
    state = load_merged_states()
    repo_name = state.get('repo')
    if not repo_name:
        logger.error('Unable to locate repo name in state.')
        return
    if push_first:
        exec_shell(f'cd {get_cwd()} && git add . && git commit -m "{release_message}" && git push -u origin {branch}')
    github_token = github_token or state.get('github_token', '')
    github = Github(login_or_token=github_token)
    repo = github.get_repo(repo_name)
    rez = repo.create_git_tag_and_release(tag = tag, tag_message = tag_message, release_name= release_name, release_message= release_message, draft = draft, prerelease = prerelease)
    logger(f'Created Release: {release_name}')
    logger(rez)

    

@repoCli.command('meta')
def display_meta(
    config_file: Optional[str] = Argument(get_cwd('metadata.yaml')),
    ):
    state = load_state()
    config_file = config_file or state.get('config_file')
    config_path = to_path(config_file)
    if not config_path.exists():
        logger.error(f'{config_file} does not exist')
        return
    config_data = '\n' + config_path.read_text()
    logger(config_data)


@stateCli.command('local')
def display_state():
    state = load_state()
    logger(state)


@stateCli.command('global')
def display_globalstate():
    state = load_global_state()
    logger(state)


@stateCli.command('merged')
def display_merged_state():
    state = load_merged_states()
    logger(state)

@stateCli.command('pypi')
def display_pypi_state():
    from pylibup.config import load_pypi_creds
    state = load_merged_states()
    creds = load_pypi_creds(state.get('pypirc_path', '~/.pypirc'))
    logger(creds)


@stateCli.command('set')
def set_state(
        states: List[str],
        global_state: bool = Option(False),
        overwrite_state: bool = Option(True),
    ):
    statevals = load_merged_states()
    for state in states:
        s = state.split('=', 1)
        key, val = s[0].strip(), s[1].strip()
        if val:
            logger.info(f'Setting {key} -> {val}. Previous: {statevals.get(key)}')
            statevals[key] = val
    if global_state:
        logger.info(f'Saving Global State: {globalstatefile.as_posix()}')
        save_global_state(overwrite_state = overwrite_state, **statevals)
    else:
        logger.info(f'Saving Local State: {statefile.as_posix()}')
        save_state(overwrite_state = overwrite_state, **statevals)


import os
from pathlib import Path
from logz import get_cls_logger

from .types import *

get_logger = get_cls_logger('pylib', 'info')

def exec_shell(cmd): return os.system(cmd)
def get_parent_path(p: str) -> Path: return Path(p).parent
def to_camelcase(string: str) -> str: return ''.join(word.capitalize() for word in string.split('_'))

def to_path(path: Union[str, Path], resolve: bool = True) -> Path:
    if isinstance(path, str): path = Path(path)
    if resolve: path.resolve()
    return path

def set_to_many(value: AnyMany) -> List[Any]:
    if not isinstance(value, list): value = [value]
    return value

def does_text_validate(text: str, include: List[str] = [], exclude: List[str] = [], exact: bool = False, **kwargs) -> bool:
    if not include and not exclude: return True
    _valid = False
    if exclude:
        for ex in exclude:
            if ((exact and ex == text)  or not exact and (ex in text or text in ex)): return False
        _valid = True
    if include:
        for inc in include:
            if ((exact and inc == text) or not exact and (inc in text or text in inc)): return True
    return _valid

def does_text_match(text: str, items: TextMany, exact: bool = False, valArgs: ValidatorArgs = None, **kwargs):
    items = set_to_many(items)
    for i in items:
        if exact and i == text or (text in i or i in text):
            if valArgs: return does_text_validate(i, exact=exact, **ValidatorArgs.dict())
            return True
    return False


__all__ = [
    'get_logger',
    'exec_shell',
    'get_parent_path',
    'to_camelcase',
    'to_path',
    'Path',
    'set_to_many',
    'does_text_match',
    'does_text_validate'
]




import os
import sys
from pathlib import Path
from setuptools import setup, find_packages

if sys.version_info.major != 3:
    raise RuntimeError("This package requires Python 3+")

version = '0.0.2'
pkg_name = 'pylibup'
gitrepo = 'trisongz/pylibup'
root = Path(__file__).parent

requirements = [
    'typer',
    'PyGithub',
    'GitPython',
    'Jinja2',
    'pyyaml',
    'pylogz',
    'requests',
]

args = {
    'packages': find_packages(include = ['pylibup', 'pylibup.*']),
    'install_requires': requirements,
    'long_description': root.joinpath('README.md').read_text(encoding='utf-8'),
    'entry_points': {
        'console_scripts': [
            'pylibup = pylibup.cli:baseCli',
            'pylib = pylibup.cli:baseCli'
        ]
    }
}

setup(
    name=pkg_name,
    version=version,
    url='https://github.com/pylibup',
    license='MIT Style',
    description='Python library Builder with CLI',
    author='Tri Songz',
    author_email='ts@growthengineai.com',
    long_description_content_type="text/markdown",
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries',
    ],
    **args
)


from .envs import *
from .utils import to_path, Path
from .types import *

class PypiCreds(BaseModel):
    name: str
    username: str
    password: str


def load_pypi_creds(path: str = '~/.pypirc'):
    fpath = to_path(Path(path).expanduser())
    if not fpath.exists(): return None
    text = fpath.read_text()
    data = text.split('\n\n')
    rez = {}
    for item in data:
        items = [i.strip() for i in item.split('\n') if i.strip()]
        name = items.pop(0)
        r = {'name': name.replace('[', '').replace(']', '')}
        for i in items:
            d = i.split(' = ')
            r[d[0].strip()] = d[1].strip()
        rez[r['name']] = PypiCreds(**r)
    return rez
    
# We assume user is authenticated to git right?
class GitConfig:
    token: str = envToStr('GITHUB_TOKEN', '')
    





from github import Github
from .utils import get_logger, to_path, Path, exec_shell
from .config import GitConfig, load_pypi_creds
from .classes import PylibConfig, get_metadata_template
from typing import Dict

logger = get_logger()

class PylibClient:
    def __init__(self, github_token: str = None, pyirc_path: str = '~/.pypirc'):
        self.github_token = github_token or GitConfig.token
        self.pyirc = load_pypi_creds(pyirc_path)
        self.github = Github(login_or_token=self.github_token)
        self.cfg: PylibConfig = None
    
    def init_cfg(self, config_file: str = None, project_name: str = None, project_dir: str = None, *args, **kwargs):
        self.cfg = PylibConfig(github=self.github, github_token = self.github_token, config_file = config_file, project_name = project_name, project_dir = project_dir, *args, **kwargs)
    
    def init(self, project_dir: str = None, name: str = None, repo_user: str = None, private: bool = True, overwrite: bool = False, **kwargs):
        working_dir = to_path(project_dir) if project_dir else Path.cwd()
        if not name: name = working_dir.stem
        elif name not in working_dir.as_posix(): working_dir.joinpath(name)
        working_dir.mkdir(exist_ok=True)
        tmpl_file = working_dir.joinpath('metadata.yaml')
        if tmpl_file.exists() and not overwrite: raise ValueError(f'{tmpl_file.as_posix()} exists but overwrite is not selected')
        tmpl = get_metadata_template(github = self.github, name = name, repo_user = repo_user, private = private, **kwargs)
        logger.info(f'Writing Metadata template to {tmpl_file.as_posix()}')
        tmpl = "## Autogenerated by Pylibup\n\n" + tmpl
        tmpl_file.write_text(tmpl)
        #logger.info
        #logger.info(tmpl)

    def build(self, config_file: str = None, project_name: str = None, project_dir: str = None, commit_msg: str = 'Initialize', overwrite: bool = False, auto_publish: bool = False, *args, **kwargs):
        self.init_cfg(config_file= config_file, project_name = project_name, project_dir = project_dir, *args, **kwargs)
        self.cfg.build(commit_msg= commit_msg, overwrite= overwrite, auto_publish= auto_publish, *args, **kwargs)
        if auto_publish:
            extra_secrets = {}
            if self.cfg.config.needs_ipyirc:
                extra_secrets['PYPI_API_TOKEN'] = self.pyirc.get('pypi')
            self.create_secrets(extra_secrets)

    def create_secrets(self, extra_secrets: Dict[str, str] = {}):
        if not self.cfg: return
        assert self.cfg.repo_exists, 'Repo has not been created yet'
        repo = self.github.get_repo(self.cfg.config.repo_path)
        secrets = self.cfg.config.get_secrets()
        if secrets:
            for key, value in secrets.items():
                rez = repo.create_secret(secret_name = key, unencrypted_value = value)
                logger.info(f'Created Secret: {key} = {rez}')
        if extra_secrets:
            for key, value in extra_secrets.items():
                rez = repo.create_secret(secret_name = key, unencrypted_value = value)
                logger.info(f'Created Secret: {key} = {rez}')

    def publish(self, commit_msg: str = "Initialize", config_file: str = None, project_name: str = None, project_dir: str = None, *args, **kwargs):
        self.init_cfg(config_file= config_file, project_name = project_name, project_dir = project_dir, *args, **kwargs)
        self.cfg.publish_repo(commit_msg = commit_msg)



    









